/********************************************************************
 * cc.c          1-NOV-1993            Y.Takeuchi (T.I.T.)          *
 *                                                                  *
 * for Soralis2.x SF-110 K2917 K3922                                *
 *                                                                  *
 *  1-NOV-1993 Y.Takeuchi    ported from SunOS4.1.x version         *
 *  5-Dec-1994 Y.Takeuchi    under development version (w/o DMA)    *
 *  26-May-1997 E.Inoue      FORCE SPARC CPU-5V version		    *
 *  26-Jul-1999 E.Inoue, K.Takahashi
 *                           FORCE SPARC CPU-50T version	    *
 ********************************************************************/
#include <sys/types.h>
#include <sys/param.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <sys/conf.h>
#include <sys/uio.h>
#include <sys/map.h>
#include <sys/debug.h>
#include <sys/modctl.h>
#include <sys/kmem.h>
#include <sys/cmn_err.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/ddi_impldefs.h>
#include <sys/ddi.h>
/* add logichouse */
#include <sys/ddidmareq.h>
#include <sys/dmaga.h>
/* end */
#include <sys/sunddi.h>

#include "cc_config.h"
#include "cc_common.h"
#include "cc.h"
#include "k2917.h"

/******************************************************
 * Data Structure                                     *
 ******************************************************/
 
/* add logichouse */
static ddi_dma_attr_t cc_dma_attr =
{
    DMA_ATTR_V0,                    /* dma_attr_version */
    (unsigned long long)0x0,        /* dma_attr_addr_lo */
    (unsigned long long)0xffffffff, /* dma_attr_addr_hi */
    (unsigned long long)0xffffffff, /* dma_attr_count_max */
    0x4,                            /* dma_attr_align */
    DEFAULT_BURSTSIZE | BURST32,    /* dma_attr_burstsizes */
    0x4,                            /* dma_attr_minxfer */
    (unsigned long long)0xffffffff, /* dma_attr_maxxfer */
    (unsigned long long)0xffffffff, /* dma_attr_seg */
    17,                             /* dma_attr_sgllen */
    512,                            /* dma_attr_granular */
    0,                              /* dma_attr_flags */
};
/* end */

static void *cc_state;

struct cc_device {
    int status;		  /* current system status */
    int cc_busy;          /* true if the device is busy */
    struct K_REG *k;      /* current pointer of registers of K2917 */
    u_short *ptr_list;	  /* current pointer to list_buffer */
    int len_list;	  /* actual length processed */
    u_short *ptr_udata_s; /* first pointer to user's buffer address */
    u_short *ptr_udata;	  /* current pointer to user's buffer address */
    int len_udata_t;	  /* total length of user's buffer */
    int len_udata;	  /* actual length of user's buffer */
    u_short *ptr_kdata;	  /* current pointer to kernel's buffer address */
    int len_kdata;	  /* actual length of kernel's buffer processed */
    int s_reg;            /* S-register of KEK-list processing */ 
    int t_reg;            /* T-register of KEK-list processing */ 
    int a_reg;            /* A-register of KEK-list processing */ 
    int event_count;      /* event counter for KEK-list processing */
    u_short camac_qx;     /* CAMAC status (CSR of K2917) */
    int branch_flag;      /* internal flag for KEK-list processing */
    int max_branch;       /* maximum number of K2917 (fixed at 1) */
    int cur_branch;       /* current branch of K2917 (fixed at 0) */        
    int cur_crate;        /* current CAMAC crate address of K3922 (0-7) */
    int interrupt;        /* interrupt source */
    u_short mode;         /* K2917 mode */
    u_short naf;          /* CAMAC NAF (N<<9)+(A<<5)+F */
    int len;              /* data length */
    int retlen;           /* transfered data length */
    struct uio *uio; 
    dev_t dev; 
    int klist;            /* flag for K2917 list processing */

    /* for solaris 2.x */
    dev_info_t *dip;
    kmutex_t mutex;
    kcondvar_t cv;
    ddi_iblock_cookie_t iblock_cookie;  /* for mutexes */
    ddi_dma_cookie_t dma_cookie;
    ddi_dma_handle_t handle;  /* for mutexes */
/* add logichouse */
    uint_t ccount;
/* end */
    int timeout_id;
    struct buf *bp;
    int executing_dma_flag;
    int cc_sys_status;    /* temporally system status storage */
/* add E.Inoue */
    struct cc_iosb iosb;
    struct cc_message message;
    u_short data_area[CC_LIST_MAX_DATA];
    u_short list_area[CC_LIST_MAX_LENGTH];
    struct klist klist1[CC_KLIST_MAX_DIR];
    int ptr_dir;
    u_short ptr_cma, now_cma, now_wc;
/* end */
/* add E.Inoue */
    int     data1;
/* end */
} ccdevice[NCC];

/**************************************************************************
 * Prototypes.                                                            *
 * These are the entry points into our driver that are called when the
 * driver is loaded, during a system call. or in response to an interrupt.
 **************************************************************************/
/*static int cc_info(dev_info_t *, ddi_info_cmd_t , void *, void **); */
static int cc_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
		 void **result);
static int cc_identify(dev_info_t *dip);
static int cc_probe(dev_info_t *dip);
static int cc_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static int cc_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);
static int cc_open(dev_t *devp, int flag, int otyp, cred_t *cred_p);
static int cc_close(dev_t dev, int flag, int otyp, cred_t *cred_p);
static int cc_write(dev_t dev, struct uio *uio, cred_t *cred_p);
static int cc_ioctl(dev_t dev, int cmd, int arg, int flag, cred_t *cred_p,
		 int *rval_p);
static int cc_strategy(struct buf *bp);
static void cc_minphys(struct buf *bp);
static u_int cc_intr();
/* By E.Inoue */
/*static void cc_timeout(); */
static void cc_timeout(caddr_t unit);
static int camac_s(int unit, u_short mode, u_short naf,
		 u_short *dat);
static int camac_b(int unit, u_short mode, u_short naf,
		 int len, int *retlen);

/**************************************************************************
 * Device driver ops vector.                                              *
 * cb_ops - character/block entry points structure.
 * When our driver is loaded or unloaded, the system calls our _init or
 * _fini routine with a modlinkage structure. The modlinkage structure
 * contains:
 * 
 *	modlinkage ->
 *		modldrv ->
 *			dev_ops ->
 *				cp_ops ->
 * cp_ops contains the normal driver entry points and roughly equivalent
 * to the cdevsw & bdevsw structures in previous releases.
 * 
 * dev_ops - device operations structure.
 * dev_ops contains, in addition to pointer to cb_ops, the routines
 * that support loading and unloading our river.
 **************************************************************************/
static struct cb_ops cc_cb_ops = {
    cc_open,		/* open */
    cc_close,		/* close */
    nodev,              /* strategy (not block device)  */
    nodev,		/* print    (no print routine)  */
    nodev,		/* dump     (no dump routine)   */
    nodev,		/* read     (no read routine; write-only device */
    cc_write,		/* write */
    cc_ioctl,		/* ioctl */
    nodev,		/* devmap   (no devmap routine) */
    nodev,		/* mmap     (no mmap routine)   */
    nodev, 		/* segmap   (no segmap routine) */
    nochpoll,		/* poll     (no chpoll routine) */
    ddi_prop_op,	/* cb_prop_op */
    0,			/* streamtab  */
    D_NEW | D_MP	/* Driver comaptibility flag */
};

struct dev_ops cc_ops = {
    DEVO_REV,		/* devo_rev, */
    0,			/* refcnt  */
    cc_getinfo,		/* info */
    cc_identify,	/* identify */
    cc_probe,		/* probe */
    cc_attach,		/* attach */
    cc_detach,		/* detach */
    nodev,		/* reset (device reset routine) */
    &cc_cb_ops,		/* driver operations */
    (struct bus_ops *)0	/* bus operations */
};


/***************************************************
 * Loadable module wrapper                         *
 ***************************************************/
extern struct mod_ops mod_driverops;

static struct modldrv modldrv = {
    &mod_driverops,
    "CAMAC driver v1.4x by Y.Takeuchi(T.I.T.)",
    &cc_ops,
};

static struct modlinkage modlinkage = {
    MODREV_1, 
    (void *)&modldrv,
    NULL,
};


/***********************************************************************
 * Module initialization routine.                                      *
 * _init, _info, and _fini support loading and unloading the driver.
 ***********************************************************************/
int _init(void)
{
    int error;

    if((error = ddi_soft_state_init(&cc_state,sizeof(struct cc_device),1)) != 0){
        return(error);
    }
    if((error = mod_install(&modlinkage)) != 0){
        ddi_soft_state_fini(&cc_state);
    }
        
    return(error);
}

int _info(struct modinfo *modinfop)
{
    return (mod_info(&modlinkage, modinfop));
}

int _fini(void)
{
    int error;

    if((error = mod_remove(&modlinkage)) == 0){
        ddi_soft_state_fini(&cc_state);
    }
    return (error);
}

/***************************************************
 * Autoconfiguration Routines                      *
 ***************************************************/
static int
cc_identify(dev_info_t *dip)
{
    char *name = ddi_get_name(dip);
    
    if (strcmp(name, "cc") == 0) {
	return (DDI_IDENTIFIED);
    } else {
	return (DDI_NOT_IDENTIFIED);
    }
}

static int
cc_probe(dev_info_t *dip)
{
    int	*reg, value;
    int instance;
    struct regspec *rp;
    register struct cc_device *cc;
/*  register struct cc_device *cc = &ccdevice[0]; */

    rp = ddi_rnumber_to_regspec(dip, 0);
    reg = (int *)0;

    instance = ddi_get_instance(dip);

    /* initialize cc_device */

    if(ddi_soft_state_zalloc(cc_state,instance) != DDI_SUCCESS){
        cmn_err(CE_WARN,"cc_probe: ddi_soft_state_zalloc failed");
        return (DDI_FAILURE);
    }

    if((cc = ddi_get_soft_state(cc_state,instance)) == NULL){
        cmn_err(CE_WARN,"cc_probe: ddi_get_soft_state failed");
        ddi_soft_state_free(cc_state,instance);
        goto failure;
    }

/*  bzero((char *)cc, sizeof(struct cc_device)); */
    cc->dip = dip;
/*  cc->k = (struct K_REG *)reg; */

    /* Map in device registers */
    if (ddi_map_regs(dip, (u_int)0, (caddr_t *)&cc->k, (off_t)0, 0)) {
	uprintf("cc_probe: unable to map registers for cc0\n");
	goto failure;
    }

    /* check K2917 */
    if (ddi_peeks(dip,(short *)&cc->k->csr,(short *)&value) == DDI_FAILURE) {
	uprintf("cc_probe: unable to find K2917 registers at 0x%X\n",
	       rp->regspec_addr);
	goto failure;
    }

    /* return */
    printf("%s\n",CC_START_MESSAGE);
    return (DDI_PROBE_SUCCESS);

failure:
    if (reg)
      ddi_unmap_regs(dip, 0, (caddr_t *)&cc->k, (off_t) 0, (off_t) 0);
    return (DDI_PROBE_FAILURE);
}

cc_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
    u_short value;
    register struct cc_device *cc;
/*  register struct cc_device *cc = &ccdevice[0]; */
/*  register struct K_REG *k = cc->k; */
    int instance;

    instance = ddi_get_instance(dip);

    if((cc = ddi_get_soft_state(cc_state,instance)) == NULL){
        cmn_err(CE_WARN,"cc_probe: ddi_get_soft_state failed");
        return (DDI_FAILURE);
    }

    /* save status to log file */
    ddi_report_dev(dip);

    /* reset & check K2917 */
    value = (u_short)CC_RST;
    ddi_pokes(dip, (short *)&cc->k->csr, (short)value);
    ddi_peeks(dip, (short *)&cc->k->csr, (short *)&value);
    if (value != 0x0080) {
/*        uprintf("cc_attach: unable to reset K2917\n"); */
        cmn_err(CE_WARN,"cc_attach: unable to reset K2917\n");
        return (DDI_FAILURE);
    }

    /* create minor node */
    if (ddi_create_minor_node(dip, ddi_get_name(dip), S_IFCHR, 0,
                              NULL, NULL) == DDI_FAILURE) {
        ddi_remove_minor_node(dip, NULL);
/*        uprintf("cc_attach: unable to create minor node\n"); */
        cmn_err(CE_WARN,"cc_attach: unable to create minor node\n");
        return (DDI_FAILURE);
    }

    /* add interrupt handller */
    if (ddi_add_intr(dip, 0, &cc->iblock_cookie,
                     (ddi_idevice_cookie_t *) NULL, cc_intr,
                     (caddr_t) instance) != DDI_SUCCESS) {
        ddi_remove_minor_node(dip, NULL);
/*        uprintf("cc_attach: unable to install high-interrupt handler\n"); */
        cmn_err(CE_WARN,"cc_attach: unable to install high-interrupt handler\n");
        return (DDI_FAILURE);
    }

    /* initialize MUTEX */
    mutex_init(&cc->mutex, "CAMAC mutex",
               MUTEX_DRIVER, (void *)&cc->iblock_cookie);

    /* Initialize the conditional variables. */
    cv_init(&cc->cv, "CAMAC cv", CV_DRIVER, NULL);

/* add logichouse */
    if (ddi_dma_alloc_handle(dip, &cc_dma_attr,DDI_DMA_SLEEP, NULL,
        &cc->handle) != DDI_SUCCESS) {
        cmn_err(CE_WARN,"cc_attach: can't alloc dma handle");
        return (DDI_FAILURE);
    }
/* end */


    return (DDI_SUCCESS);
}


static int
cc_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{ 
    register struct cc_device *cc;
/*  register struct cc_device *cc = &ccdevice[0]; */
    int instance;

    /* deallocate resources from attach */

    instance = ddi_get_instance(dip);

    cc = ddi_get_soft_state(cc_state,instance);
    
/* add logichouse */
    ddi_dma_free_handle(&cc->handle);
/* end */

    ddi_remove_minor_node(dip, NULL);
    ddi_remove_intr(dip, 0, cc->iblock_cookie);
    mutex_destroy(&cc->mutex);
    cv_destroy (&cc->cv);
    ddi_soft_state_free(cc_state,instance);

    return (DDI_SUCCESS);
}

/*
 * Given the device number return the devinfo pointer.
 */
/* ARGSUSED */
static int
cc_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg, void **result)
{
/*  register dev_t dev = (dev_t) arg; */
    register int instance, error;
/*  register struct cc_device *cc = &ccdevice[0]; */
    register struct cc_device *cc;

    switch (infocmd) {
      case DDI_INFO_DEVT2DEVINFO:
        instance = getminor((dev_t)arg);
        if((cc = ddi_get_soft_state(cc_state,instance)) != NULL){
	    *result = (void *)cc->dip;
	    error = DDI_SUCCESS;
        }
	break;
      case DDI_INFO_DEVT2INSTANCE:
/*	instance = 0; */
        instance = getminor((dev_t)arg);
	*result = (void *)instance;
	error = DDI_SUCCESS;
	break;
      default:
        *result = NULL;
	error = DDI_FAILURE;
    }
    return (error);
}

/*****************************************************************************/
#define CCHPRI (PZERO - 1)
#define CCLPRI (PZERO + 1)

extern int hz;

/******************************************************
 * Internal routines for CAMAC list processing        *
 ******************************************************/
extern int cc_stop(),
	cc_read16(),  cc_read24(),  cc_read16_a(),  cc_read24_a(),
	cc_write16(), cc_write24(), cc_write16_a(), cc_write24_a(),
	cc_ndt(), cc_setcrate_a(), cc_setcrate(), cc_genz(),
	cc_genc(), cc_remi(), cc_seti(), cc_enbint(), cc_dsbint(),
	cc_qwait(),
	cc_qs16(), cc_qs24(), cc_qs16_a(), cc_qs24_a(),
	cc_as16(), cc_as24(), cc_as16_a(), cc_as24_a(),
	cc_igq16(), cc_igq24(), cc_igq16_a(), cc_igq24_a(),
	cc_move_as(), cc_move_at(), cc_move_sa(), cc_move_st(),
	cc_move_ta(), cc_move_ts(),
	cc_move32_a(), cc_move32_s(), cc_move32_t(), cc_move16_a_ptr(),
	cc_move32_a_ptr(), cc_move16_ptr(), cc_move32_ptr(),
	cc_move16_a_head(), cc_move16_head_a(), cc_move16_head(),
	cc_load_ptr(), cc_load_remain(), cc_load_timer(), 
	cc_load_count(), cc_load_qx(),
	cc_head(), cc_exchg_at(), cc_swap_a(),
	cc_bis32_at(), cc_bis32_ta(), cc_bis32_a(), cc_bis32_t(),
	cc_bic32_at(), cc_bic32_ta(), cc_bic32_a(), cc_bic32_t(),
	cc_add32_at(), cc_add32_ta(), cc_add32_a(), cc_add32_t(),
	cc_sub32_at(), cc_sub32_ta(), cc_sub32_a(), cc_sub32_t(),
	cc_mul32_at(), cc_mul32_ta(), cc_mul32_a(), cc_mul32_t(),
	cc_div32_at(), cc_div32_ta(), cc_div32_a(), cc_div32_t(),
	cc_waitint(), cc_waitevent(),
	cc_br(), cc_beq(), cc_bne(), cc_bgt(), cc_bge(),
        cc_blt(), cc_ble(), cc_bbc(), cc_bbs(),
        kl_stop(),
	kl_read16(),  kl_read24(), kl_write16(), kl_write24(), kl_ndt(),
	kl_qs16(), kl_qs24(), kl_as16(), kl_as24(), kl_igq16(), kl_igq24();

/* for KEK-list processing */  
int (*instruction_call)();
int (*instruction[0x80])() = {
/* 0 */			cc_stop,
/* 1,2,3,4 */		cc_read16,  cc_read24,  cc_read16_a,  cc_read24_a,
/* 5,6,7,8 */		cc_write16, cc_write24, cc_write16_a, cc_write24_a,
/* 9,A,B,C */		cc_ndt, cc_setcrate_a, cc_setcrate, cc_genz,
/* D,E,F,10,11 */	cc_genc, cc_remi, cc_seti, cc_enbint, cc_dsbint,
/* 12 */		cc_qwait,
/* 13,14,15,16 */	cc_qs16, cc_qs24, cc_qs16_a, cc_qs24_a,
/* 17,18,19,1A */	cc_as16, cc_as24, cc_as16_a, cc_as24_a,
/* 1B,1C,1D,1E */	cc_igq16, cc_igq24, cc_igq16_a, cc_igq24_a,
/* 1F - 22 */		cc_stop, cc_stop, cc_stop, cc_stop,
/* 23,24,25,26 */	cc_move_as, cc_move_at, cc_move_sa, cc_move_st,
/* 27,28 */		cc_move_ta, cc_move_ts,
/* 29,2A,2B,2C */	cc_move32_a, cc_move32_s, cc_move32_t, cc_move16_a_ptr,
/* 2D,2E,2F */		cc_move32_a_ptr, cc_move16_ptr, cc_move32_ptr,
/* 30,31,32 */		cc_move16_a_head, cc_move16_head_a, cc_move16_head,
/* 33,34,35 */		cc_load_ptr, cc_load_remain, cc_load_timer, 
/* 36,37 */		cc_load_count, cc_load_qx,
/* 38,39,3A */		cc_head, cc_exchg_at, cc_swap_a,
/* 3B,3C,3D,3E */	cc_bis32_at, cc_bis32_ta, cc_bis32_a, cc_bis32_t,
/* 3F,40,41,42 */	cc_bic32_at, cc_bic32_ta, cc_bic32_a, cc_bic32_t,
/* 43,44,45,46 */	cc_add32_at, cc_add32_ta, cc_add32_a, cc_add32_t,
/* 47,48,49,4A */	cc_sub32_at, cc_sub32_ta, cc_sub32_a, cc_sub32_t,
/* 4B,4C,4D,4E */	cc_mul32_at, cc_mul32_ta, cc_mul32_a, cc_mul32_t,
/* 4F,50,51,52 */	cc_div32_at, cc_div32_ta, cc_div32_a, cc_div32_t,
/* 53,54 */		cc_waitint, cc_waitevent,
/* 55,56,57,58,59 */	cc_br, cc_beq, cc_bne, cc_bgt, cc_bge,
/* 5A,5B,5C,5D */	cc_blt,cc_ble, cc_bbc, cc_bbs,
/* 5E,5F */		cc_stop, cc_stop,

/* for K2917 list processing */
/* 0x60 + */
/* 0 */			kl_stop,
/* 1,2,3,4 */		kl_read16,  kl_read24,  kl_stop, kl_stop,
/* 5,6,7,8 */		kl_write16, kl_write24, kl_stop, kl_stop,
/* 9,A,B,C */		kl_ndt, kl_stop, kl_stop, kl_stop,
/* D,E,F,10,11 */	kl_stop, kl_stop, kl_stop, kl_stop, kl_stop,
/* 12 */		kl_stop,
/* 13,14,15,16 */	kl_qs16, kl_qs24, kl_stop, kl_stop,
/* 17,18,19,1A */	kl_as16, kl_as24, kl_stop, kl_stop,
/* 1B,1C,1D,1E */	kl_igq16, kl_igq24, kl_stop, kl_stop

};

/******************************************************
 * Internal data areas                                *
 ******************************************************/
/*
static struct cc_iosb iosb;
static struct cc_message message;
static u_short data_area[CC_LIST_MAX_DATA];
static u_short list_area[CC_LIST_MAX_LENGTH];
static struct klist klist[CC_KLIST_MAX_DIR];
static int ptr_dir;
static u_short ptr_cma, now_cma, now_wc;
*/
/* static int cc_sys_status;           temporally system status storage */

union swap {
    long longs;
    short shorts[2];
} swap;


u_short intrpri = 4;
u_short intrvec = 0xff;


/******************************************************
 * ccopen                                             *
 ******************************************************/
static int
cc_open(dev_t *devp, int flag, int otyp, cred_t *cred_p)
{
    register int unit;
/*  register struct cc_device *cc = &ccdevice[0]; */
    register struct cc_device *cc;
    void bzero();
/* by E.Inoue */
/*  register struct K_REG *k = cc->k; */
/* */

   unit = getminor(*devp);
   cc = ddi_get_soft_state(cc_state,unit);

/*
    if (unit >= NCC || !md || md->md_alive == 0)
        return (ENXIO);
*/

    if (cc->cc_busy == CC_BUSY)
        return (EBUSY);

    /* initialize ccdevice */
    cc->max_branch = CC_K_MAX_BRANCH;

    /* set busy flag */
    cc->cc_busy = CC_BUSY;

/* by E.Inoue */
    cc->k->csr = CC_RST;                         /* K2917 Reset */
/* */

    /* set interrupt registers of K2917 */
    cc->k->lamc = (u_short)(CC_INT_AUTO_CLEAR | intrpri);
    cc->k->donc = (u_short)(CC_INT_AUTO_CLEAR | intrpri);
    cc->k->empc = (u_short)(CC_INT_AUTO_CLEAR | intrpri);
    cc->k->aboc = (u_short)(CC_INT_AUTO_CLEAR | intrpri);
    cc->k->lamv = (u_short)intrvec;
    cc->k->donv = (u_short)intrvec;
    cc->k->empv = (u_short)intrvec;
    cc->k->abov = (u_short)intrvec;

    return 0;
}

/******************************************************
 * ccclose                                            *
 ******************************************************/
static int
cc_close(dev_t dev, int flag, int otyp, cred_t *cred_p)
{
/*  register struct cc_device *cc = &ccdevice[0]; */
    register struct cc_device *cc;
    minor_t unit;

    unit = getminor(dev);

    cc = ddi_get_soft_state(cc_state,unit);
/*
    if (unit >= NCC)
        return ENXIO;
*/
    /* clear busy flag */
    cc->cc_busy = 0;
    return 0;
}

/******************************************************
 * ccwrite                                            *
 ******************************************************/
static int
cc_write(dev_t dev, struct uio *uio, cred_t *cred_p)
{
/*  register struct cc_device *cc = &ccdevice[0]; */
    register struct cc_device *cc;
    register struct iovec *iov = uio->uio_iov;
    register u_short mode, naf;
    register int len, code, step_count;
    struct uio   uio_save, *uios = &uio_save;
    struct iovec iov_save, *iovs = &iov_save;
    int retlen;
    char klname[9];
    minor_t unit;

    unit = getminor(dev);
    cc = ddi_get_soft_state(cc_state,unit);

/*
    if (unit >= NCC)
        return ENXIO;
*/
    cc->dev = dev;
    cc->uio = uio;
    cc->ptr_kdata = cc->data_area;
    cc->len_kdata = 0;

/*
    if (uiomove(&message, sizeof(struct cc_message), UIO_WRITE, uio))
        return EFAULT;
*/
    copyin((caddr_t)iov->iov_base, (caddr_t)&cc->message, sizeof(cc->message));

    cc->status = 0;
    cc->cc_sys_status = 0;
    cc->mode = mode = cc->message.mode;
    cc->naf  = naf  = cc->message.naf;

    /* save iov, uio */
    iovs->iov_base   = iov->iov_base;
    iovs->iov_len    = iov->iov_len;
    uios->uio_iovcnt = uio->uio_iovcnt;
    uios->uio_offset = uio->uio_offset;
    uios->uio_segflg = uio->uio_segflg;
    uios->uio_resid  = uio->uio_resid;

    switch (cc->message.command) {

	/*************************
	 * CAMAC single action   *
	 *************************/
      case CC_CMD_DOSINGLE:
	switch (naf & 0x0018) {
	  case 0x0000:                                  /* CAMAC read */
	    camac_s(unit, mode, naf, cc->data_area);
	    if ((mode & CC_BIT16) == 0)
	        copyout((caddr_t)cc->data_area, (caddr_t)cc->message.ptr_data,
			sizeof(u_short) * 2);
	    else
	        copyout((caddr_t)cc->data_area, (caddr_t)cc->message.ptr_data,
			sizeof(u_short));
	    break;
	  case 0x0010:                                  /* CAMAC write */
	    if ((mode & CC_BIT16) == 0)
	        copyin((caddr_t)cc->message.ptr_data, (caddr_t)cc->data_area, 
		       sizeof(u_short) * 2);
	    else
	        copyin((caddr_t)cc->message.ptr_data, (caddr_t)cc->data_area,
		       sizeof(u_short));
	    camac_s(unit, mode, naf, cc->data_area);
	    break;
	  default:                                      /* NDT */
	    camac_s(unit, mode, naf, cc->data_area);
	    break;
	}
	copyout((caddr_t)((u_short *)&cc->camac_qx),
		(caddr_t)cc->message.ptr_qx, 2);
        break;

	/*************************
	 * CAMAC block action    *
	 *************************/
      case CC_CMD_DOBLOCK:

	iov++;
	uio->uio_iov++;
	uio->uio_iovcnt--;

	cc->ptr_udata = (u_short *)iov->iov_base;
	cc->len_udata = 0;
	cc->len_udata_t = len = iov->iov_len / sizeof(u_short);

	if ((mode & CC_BIT16) == 0)
	    len /= 2;

	switch (naf & 0x0018) {
	  case 0x0000:                                  /* CAMAC read */
	    camac_b(unit, mode, naf, len, &retlen);
	    break;
	  case 0x0010:                                  /* CAMAC write */
	    camac_b(unit, mode, naf, len, &retlen);
	    break;
	  default:                                      /* NDT */
	    camac_b(unit, mode, naf, len, &retlen);
	    break;
	}
	cc->iosb.status = cc->status;
	cc->iosb.ret_length = retlen;
	cc->iosb.s_reg = cc->s_reg;
	cc->iosb.devinfo = cc->camac_qx + (cc->camac_qx << 16);
	copyout((caddr_t)&cc->iosb, (caddr_t)cc->message.ptr_iosb,
		sizeof(struct cc_iosb));
        break;

	/*************************
	 * CAMAC list processing *
	 *************************/
      case CC_CMD_LOADLIST:
      case CC_CMD_LOADDOLIST:

	iov++;
	uio->uio_iov++;
	uio->uio_iovcnt--;
	cc->ptr_list = (u_short *)iov->iov_base;
	cc->len_list = iov->iov_len / sizeof(u_short);
	copyin((caddr_t)cc->ptr_list, (caddr_t)cc->list_area, 
	       cc->len_list * sizeof(u_short));
	if (cc->message.command == CC_CMD_LOADLIST)
	    break;

      case CC_CMD_DOLIST:
	iov++;
	uio->uio_iov++;
	uio->uio_iovcnt--;
	copyin((caddr_t)cc->message.ptr_iosb, (caddr_t)&cc->iosb,
	       sizeof(struct cc_iosb));
	cc->ptr_udata_s = cc->ptr_udata = (u_short *)iov->iov_base;
	cc->len_udata_t = iov->iov_len / sizeof(u_short);
	cc->len_udata = 0;
	cc->ptr_list = cc->list_area;	/* initialize list pointer */
	cc->s_reg = cc->iosb.s_reg;

	step_count = 0;
	while (cc->status == CC_STA_OK &&
	       cc->cc_sys_status == CC_STA_OK &&
	       step_count++ < CC_LIST_MAX_STEP &&
	       cc->ptr_list <= (cc->list_area + cc->len_list)) {
	    code = 0x00FF & cc->ptr_list[0];
	    instruction_call = instruction[code];
	    cc->status = (*instruction_call)(cc);
	    next_instruction(cc);
	}
	if (step_count >= CC_LIST_MAX_STEP)
	    cc->status = CC_STA_LIST_OVERSTEP;
	flush_buf(cc);
	cc->iosb.status = cc->status;
	cc->iosb.ret_length = cc->len_udata;
	cc->iosb.s_reg = cc->s_reg;
	cc->iosb.devinfo = cc->camac_qx + (cc->camac_qx << 16);
	copyout((caddr_t)&cc->iosb, (caddr_t)cc->message.ptr_iosb,
				 sizeof(struct cc_iosb));
	break;

	/*************************
	 * K2917 list processing *
	 *************************/
      case CC_CMD_ADDKLIST:

	copyin((caddr_t)cc->message.klname, (caddr_t)klname, sizeof(klname));
	iov++;
	uio->uio_iov++;
	uio->uio_iovcnt--;
	cc->ptr_list = (u_short *)iov->iov_base;
	cc->len_list = iov->iov_len / sizeof(u_short);
	copyin((caddr_t)cc->ptr_list, (caddr_t)cc->list_area,
	       cc->len_list * sizeof(u_short));
	kladd(cc, klname);
	break;

      case CC_CMD_DELKLIST:

	copyin((caddr_t)cc->message.klname, (caddr_t)klname, sizeof(klname));
	kldel(klname);
	break;

      case CC_CMD_EXEKLIST:

	copyin((caddr_t)cc->message.klname, (caddr_t)klname, sizeof(klname));
	copyin((caddr_t)cc->message.ptr_iosb, (caddr_t)&cc->iosb,
	       sizeof(struct cc_iosb));
	iov++;
	uio->uio_iov++;
	uio->uio_iovcnt--;
	cc->ptr_udata_s = cc->ptr_udata = (u_short *)iov->iov_base;
	cc->len_udata_t = iov->iov_len / sizeof(u_short);
	cc->len_udata = 0;
	klexe(cc, klname);
	cc->iosb.status = cc->status;
	cc->iosb.ret_length = cc->retlen;
	cc->iosb.s_reg = cc->s_reg;
	cc->iosb.devinfo = cc->camac_qx + (cc->camac_qx << 16);
	copyout((caddr_t)&cc->iosb, (caddr_t)cc->message.ptr_iosb,
				 sizeof(struct cc_iosb));
	break;

      case CC_CMD_SAVEKLIST:

	klsave(cc, cc->data_area);
	copyout((caddr_t)cc->data_area,(caddr_t)cc->message.ptr_iosb,
		(cc->list_area[0] + 1) * 2);
	break;

      case CC_CMD_LOADKLIST:
	copyin((caddr_t)(u_short *)cc->message.ptr_iosb, (caddr_t)cc->data_area, 2);
	copyin((caddr_t)((u_short *)cc->message.ptr_iosb + 1),
	       (caddr_t)&cc->data_area[1], cc->data_area[0] * 2);

	klload(cc, cc->data_area);
	break;

      default:
	break;
    }

    /* restore iov, uio */
    iov->iov_base   = iovs->iov_base;
    iov->iov_len    = iovs->iov_len;
    uio->uio_iovcnt = uios->uio_iovcnt;
    uio->uio_offset = uios->uio_offset;
    uio->uio_segflg = uios->uio_segflg;
    uio->uio_resid  = uios->uio_resid;

    return cc->cc_sys_status;
}


/******************************************************
 * cctimeout                                          *
 ******************************************************/
static void
cc_timeout(caddr_t unit)
{
/*  register struct cc_device *cc = &ccdevice[0]; */
    register struct cc_device *cc;

/*    cc = ddi_get_soft_state(cc_state,unit); */
    cc = ddi_get_soft_state(cc_state,(int)*unit);

    mutex_enter(&cc->mutex);			/* start MUTEX */

    /* free DMA resources */
    if (cc->executing_dma_flag != 0) {
/* chenge logichouse */
/*
	ddi_dma_free(cc->handle);
*/
    ddi_dma_unbind_handle(cc->handle);
/* end */
        cc->cc_sys_status = cc->bp->b_error;
	biodone(cc->bp);
	cc->executing_dma_flag = 0;
    }

    /* set timeout */
    cc->interrupt |= CC_INT_TIMEOUT;
    cv_signal(&cc->cv);

    mutex_exit(&cc->mutex);			/* end MUTEX */

    return;
}


/******************************************************
 * cc_intr                                            *
 ******************************************************/
static u_int
cc_intr(int unit)
{
/*  register struct cc_device *cc = &ccdevice[0]; */
    register struct cc_device *cc;
/*  register struct K_REG *k = cc->k; */

    cc = ddi_get_soft_state(cc_state,unit);

    mutex_enter(&cc->mutex);			/* start MUTEX */

    /* check K2917 register */
    if ((cc->k->csr & CC_LAM) != 0)
	cc->interrupt |= CC_INT_LAM;
    if ((cc->k->csr & CC_DONE) != 0)
	cc->interrupt |= CC_INT_DONE;
    if ((cc->k->empc & CC_INT_ENABLE) == 0)
	cc->interrupt |= CC_INT_EMPTY;
    if ((cc->k->aboc & CC_INT_ENABLE) == 0)
	cc->interrupt |= CC_INT_ABORT;

    /* free DMA resources */
    if (cc->executing_dma_flag != 0) {

/* change logichouse */
/*
	ddi_dma_free(cc->handle);
*/
    ddi_dma_unbind_handle(cc->handle);
/* end */
        cc->cc_sys_status = cc->bp->b_error;
	biodone(cc->bp);
	cc->executing_dma_flag = 0;
    }

    if (cc->interrupt == 0) {
	/* reject interrupt, then return */
        mutex_exit(&cc->mutex);			/* end MUTEX */
        return DDI_INTR_UNCLAIMED;
    }
    else {
	/* accept interrupt, then wake up waiting process, then return */
	untimeout(cc->timeout_id);
	cv_signal(&cc->cv);
	mutex_exit(&cc->mutex);			/* end MUTEX */
	return DDI_INTR_CLAIMED;
    }
}


/******************************************************
 * ccioctl                                            *
 ******************************************************/
static int
cc_ioctl(dev_t dev, int cmd, int arg, int flag, cred_t *cred_p, int *rval_p)
{
/*  register struct cc_device *cc = &ccdevice[0]; */
    register struct cc_device *cc;
/*  register struct K_REG *k = cc->k; */
    u_short sdat;
/*  int     idat, *data = (int *)arg; */
/* E.Inoue 
    int     idat, data;
  end */
    int     idat;
    int s;
    minor_t unit;

    unit = getminor(dev);
    cc = ddi_get_soft_state(cc_state,unit);

/* E.Inoue
    copyin((caddr_t)arg,(caddr_t)&data,sizeof(data));
    copyin((caddr_t)arg,(caddr_t)&cc->data1,sizeof(cc->data1));
*/
    ddi_copyin((caddr_t)arg,(caddr_t)&cc->data1,sizeof(cc->data1),flag);

    switch (cmd) {
      case CCIOC_CHK_BRANCH:
/* E.Inoue
         data = cc->max_branch;
*/
         cc->data1 = cc->max_branch;
         break;
      case CCIOC_SET_BRANCH:
/* E.inoue
         if (data >= 0 && data < cc->max_branch)
           cc->cur_branch = data;
*/
         if (cc->data1 >= 0 && cc->data1 < cc->max_branch)
           cc->cur_branch = cc->data1;
         break;
      case CCIOC_SET_CRATE:
/* E.Inoue
        if (data >= 0 && data < MAX_CRATE)
          cc->cur_crate = data;
*/
        if (cc->data1 >= 0 && cc->data1 < MAX_CRATE)
          cc->cur_crate = cc->data1;
         break;
      case CCIOC_WAIT_LAM :
       mutex_enter(&cc->mutex);                        /* start MUTEX */
         cc->interrupt = 0;
         cc->k->lamc = CC_INT_AUTO_CLEAR | CC_INT_ENABLE | intrpri;

 /* lock out clock */
/* by E.Inoue
 s = spl5();
*/
 if (cc->interrupt & CC_INT_LAM) {
     cc->interrupt &= ~CC_INT_LAM;
     return 0;
 }

/* mutex_enter(&cc->mutex);  */                      /* start MUTEX */
/* E.Inoue
       cc->timeout_id = timeout(cc_timeout, (caddr_t)&unit, data * hz);
*/
       cc->timeout_id = timeout(cc_timeout, (caddr_t)&unit, cc->data1 * hz);
/*       mutex_enter(&cc->mutex);       */                 /* start MUTEX */

 /* wait interrupt */
/*
       cv_timedwait(&cc->cv, &cc->mutex, timeout);
       if (cc->interrupt == 0)
          cc->interrupt |= CC_INT_TIMEOUT;
*/
       if (cv_wait_sig(&cc->cv, &cc->mutex) == 0) {
           untimeout(cc->timeout_id);
           mutex_exit(&cc->mutex);                     /* end MUTEX */
           return EINTR;
       }
       mutex_exit(&cc->mutex);                         /* end MUTEX */

/* by E.Inoue
       splx(s);
*/

       if( cc->interrupt & CC_INT_LAM ) {
           cc->interrupt &= ~CC_INT_LAM;
       } else if( cc->interrupt & CC_INT_TIMEOUT ) {
           cc->interrupt &= ~CC_INT_TIMEOUT;
            return CC_STA_SINGLE_TIMEOUT;   /* set value to "errno" */
       }
       break;

      case CCIOC_ENABLE_LAM :
       mutex_enter(&cc->mutex);                        /* start MUTEX */
       ddi_copyin((caddr_t)arg,(caddr_t)&cc->data1,sizeof(cc->data1),flag);
/* E.Inoue
        camac_s(unit, CC_BIT24, NAF(30, 13, 17), (u_short *)data); 
*/
        camac_s(unit, CC_BIT24, NAF(30, 13, 17), (u_short *)&cc->data1);
        camac_s(unit, CC_BIT16, NAF(30,  0,  1), &sdat);
        sdat |= 0x100;
        camac_s(unit, CC_BIT16, NAF(30,  0, 17), &sdat);
        mutex_exit(&cc->mutex);                         /* end MUTEX */
        break;

      case CCIOC_DISABLE_LAM :
        cc->k->lamc = CC_INT_AUTO_CLEAR | intrpri;
        idat = 0;
        camac_s(unit, CC_BIT24, NAF(30, 13, 17), (u_short *)&idat);
        camac_s(unit, CC_BIT16, NAF(30,  0,  1), &sdat);
        sdat &= ~0x100;
        camac_s(unit, CC_BIT16, NAF(30,  0, 17), &sdat);
        break;
      case CCIOC_REG_DUMP :
/*
        data[0] = k->cser;
        data[1] = k->docr;
        data[2] = k->sccr;
        data[3] = k->mtc;
        data[4] = k->machi;
        data[5] = k->maclo;
        data[6] = k->lamc;
        data[7] = k->donc;
        data[8] = k->empc;
        data[9] = k->aboc;
        data[10] = k->lamv;
        data[11] = k->donv;
        data[12] = k->empv;
        data[13] = k->abov;
        data[14] = k->amr;
        data[15] = k->cma;
        data[16] = k->cwc;
        data[17] = k->srr;
        data[18] = k->csr;
*/
        break;
      case CCIOC_RESET :
        cc->k->csr = CC_RST;
/* by E.Inoue  from >>> */
    /* set interrupt registers of K2917 */
        cc->k->lamc = (u_short)(CC_INT_AUTO_CLEAR | intrpri);
        cc->k->donc = (u_short)(CC_INT_AUTO_CLEAR | intrpri);
        cc->k->empc = (u_short)(CC_INT_AUTO_CLEAR | intrpri);
        cc->k->aboc = (u_short)(CC_INT_AUTO_CLEAR | intrpri);
        cc->k->lamv = (u_short)intrvec;
        cc->k->donv = (u_short)intrvec;
        cc->k->empv = (u_short)intrvec;
        cc->k->abov = (u_short)intrvec;
/* to <<<  */
        break;
      case CCIOC_READ_STATUS :
/*
        data[0] = k->cser;
        data[1] = k->docr;
        data[2] = k->sccr;
        data[3] = k->mtc;
        data[4] = k->machi;
        data[5] = k->maclo;
        data[6] = k->lamc;
        data[7] = k->donc;
        data[8] = k->empc;
        data[9] = k->aboc;
        data[10] = k->lamv;
        data[11] = k->donv;
        data[12] = k->empv;
        data[13] = k->abov;
        data[14] = k->amr;
        data[15] = k->cma;
        data[16] = k->cwc;
        data[17] = k->srr;
        data[18] = k->csr;
*/
/* k->csr = CC_RST; */
       break;

      case CCIOC_KINIT :
        cc->ptr_dir = 0;
        cc->ptr_cma = CC_KLIST_CMAINIT;
        break;


      default :
        break;
    }

    return 0;
}



/********************************************************************
 * cclist.c     30-NOV-1991            Y.Takeuchi (TIT)             *
 *                                                                  *
 * original: cc.c on SUN4 by Y.Yasu (KEK)                           *
 * for SUN-SPARC SF-VME K2917 K3922                                 *
 *                                                                  *
 * 24-JUN-1992 Y.Takeuchi    fix bugs of cc_swap_a(),               *
 *                           cc_head(), cc_move**_head().           *
 * 13-JUL-1992 Y.Takeuchi    delete "NCC", modify "splx" etc.       *
 * 30-NOV-1992 Y.Takeuchi    modify waitint(), TIMEOUT_VALUE etc.   *
 * 25-JAN-1993 Y.Takeuchi    fix bug of ptr_prev_head[?]            *
 *  5-MAR-1993 Y.Takeuchi    modify for unified UNIX camlib         *
 * 17-MAR-1993 Y.Takeuchi    modify for re-execution                *
 * 30-MAR-1993 Y.Takeuchi    fix bugs of read24(), read16() (NOQ))  *
 ********************************************************************/
#define Tolong(i)  (cc->ptr_list[(i)] + (cc->ptr_list[(i) + 1] << 16))

/* gloval variables for list processing */
static u_short *ptr_prev_head; 


next_instruction(cc)
struct cc_device *cc;
{
    if (cc->branch_flag)
	cc->branch_flag = 0;
    else
	cc->ptr_list += (0xFF00 & *cc->ptr_list) >> 8;
    return 0;
}


flush_buf(cc)
struct cc_device *cc;
{
    cc->cc_sys_status = copyout((caddr_t)cc->data_area, (caddr_t)cc->ptr_udata,
			    cc->len_kdata * sizeof(u_short));
    cc->ptr_udata += cc->len_kdata;
    cc->len_udata += cc->len_kdata;
    cc->len_kdata = 0;
    cc->ptr_kdata = cc->data_area;
}


check_buf(cc, wc)
struct cc_device *cc;
int wc;                                     /* wc: word counts */
{
    if (cc->len_kdata + wc >= CC_LIST_MAX_DATA &&
	cc->len_kdata + wc < cc->len_udata_t - cc->len_udata)
	    flush_buf(cc);
    if (cc->len_kdata + wc >= cc->len_udata_t - cc->len_udata) {
        return CC_STA_LIST_EMPTY;
    }
    else
	return 0;
}


cc_stop(cc)
struct cc_device *cc;
{
    return CC_STA_LIST_STOP;
}


cc_read16(cc)
struct cc_device *cc;
{
    register int status;

    if (status = check_buf(cc, 1))
	return status;
    if (status = camac_s(getminor(cc->dev), CC_BIT16, cc->ptr_list[1], cc->ptr_kdata))
        return status;
    /* increment data pointer (if CAMAC Q=1) */
    if ((cc->camac_qx & CC_NOQ) == 0) {
	cc->ptr_kdata += 1;
	cc->len_kdata += 1;
    }
    return 0;
}


cc_read24(cc)
struct cc_device *cc;
{
    register int status;

    if (status = check_buf(cc, 2))
	return status;
    if (status = camac_s(getminor(cc->dev), CC_BIT24, cc->ptr_list[1], cc->ptr_kdata))
        return status;
    /* increment data pointer (if CAMAC Q=1) */
    if ((cc->camac_qx & CC_NOQ) == 0) {
	cc->ptr_kdata += 2;
	cc->len_kdata += 2;
    }
    return 0;
}


cc_read16_a(cc)
struct cc_device *cc;
{
    cc->a_reg = 0;
    return camac_s(getminor(cc->dev), CC_BIT16, cc->ptr_list[1], ((u_short *)&cc->a_reg + 1));
}


cc_read24_a(cc)
struct cc_device *cc;
{
    return camac_s(getminor(cc->dev), CC_BIT24, cc->ptr_list[1], (u_short *)&cc->a_reg);
}


cc_write16(cc)
struct cc_device *cc;
{
    return camac_s(getminor(cc->dev), CC_BIT16, cc->ptr_list[1], &cc->ptr_list[2]);
}


cc_write24(cc)
struct cc_device *cc;
{
    int dat;

    dat = Tolong(2);
    return camac_s(getminor(cc->dev), CC_BIT24, cc->ptr_list[1], (u_short *)&dat);
}


cc_write16_a(cc)
struct cc_device *cc;
{
    return camac_s(getminor(cc->dev), CC_BIT16, cc->ptr_list[1], ((u_short *)&cc->a_reg + 1));
}


cc_write24_a(cc)
struct cc_device *cc;
{
    return camac_s(getminor(cc->dev), CC_BIT24, cc->ptr_list[1], (u_short *)&cc->a_reg);
}


cc_ndt(cc)
struct cc_device *cc;
{
    u_short dummy;

    return camac_s(getminor(cc->dev), CC_BIT16, cc->ptr_list[1], &dummy);
}


cc_setcrate_a(cc)
struct cc_device *cc;
{
    if (cc->a_reg >= 0 && cc->a_reg < MAX_CRATE)
        cc->cur_crate = cc->a_reg;
    return 0;
}


cc_setcrate(cc)
struct cc_device *cc;
{
    if (cc->ptr_list[1] < MAX_CRATE)
        cc->cur_crate = cc->ptr_list[1];
    return 0;
}


cc_genz(cc)
struct cc_device *cc;
{
    register int status;
    int dat;

    if (status = camac_s(getminor(cc->dev), CC_BIT24, NAF(30, 0, 1), (u_short *)&dat))
	return status;
    dat |= 1;
    return camac_s(getminor(cc->dev), CC_BIT24, NAF(30, 0, 17), (u_short *)&dat);
}


cc_genc(cc)
struct cc_device *cc;
{
    register int status;
    int dat;

    if (status = camac_s(getminor(cc->dev), CC_BIT24, NAF(30, 0, 1), (u_short *)&dat))
	return status;
    dat |= 2;
    return camac_s(getminor(cc->dev), CC_BIT24, NAF(30, 0, 17), (u_short *)&dat);
}


cc_remi(cc)
struct cc_device *cc;
{
    register int status;
    int dat;

    if (status = camac_s(getminor(cc->dev), CC_BIT24, NAF(30, 0, 1), (u_short *)&dat))
	return status;
    dat &= ~4;
    return camac_s(getminor(cc->dev), CC_BIT24, NAF(30, 0, 17), (u_short *)&dat);
}


cc_seti(cc)
struct cc_device *cc;
{
    register int status;
    int dat;

    if (status = camac_s(getminor(cc->dev), CC_BIT24, NAF(30, 0, 1), (u_short *)&dat))
	return status;
    dat |= 4;
    return camac_s(getminor(cc->dev), CC_BIT24, NAF(30, 0, 17), (u_short *)&dat);
}


cc_enbint(cc)
struct cc_device *cc;
{
    register int status;
    int dat;

#ifdef CC_SPARC_LIST_EXTENSION
    dat = cc->a_reg;
#else
    dat = 0xFFFFFF;                          /* mask */
#endif
    if (status = camac_s(getminor(cc->dev), CC_BIT24, NAF(30, 13, 17), (u_short *)&dat))
	return status;
    if (status = camac_s(getminor(cc->dev), CC_BIT24, NAF(30, 0, 1), (u_short *)&dat))
	return status;
    dat |= 0x0100;
    return camac_s(getminor(cc->dev), CC_BIT24, NAF(30, 0, 17), (u_short *)&dat);
}


cc_dsbint(cc)
struct cc_device *cc;
{
    register int status;
    int dat;

    /* disable LAM */
    cc->k->lamc = CC_INT_AUTO_CLEAR | intrpri;

    dat = 0x0;                              /* mask */
    if (status = camac_s(getminor(cc->dev), CC_BIT24, NAF(30, 13, 17), (u_short *)&dat))
	return status;
    if (status = camac_s(getminor(cc->dev), CC_BIT24, NAF(30, 0, 1), (u_short *)&dat))
	return status;
    dat &= ~0x0100;
    return camac_s(getminor(cc->dev), CC_BIT24, NAF(30, 0, 17), (u_short *)&dat);
}


cc_qwait(cc)
struct cc_device *cc;
{
    register int status, i;
    int dat;

    cc->camac_qx = status = 0;
    for (i=0; status==0 && (cc->camac_qx & CC_NOQ)!=0 && i<cc->ptr_list[2];i++)
	status = camac_s(getminor(cc->dev), CC_BIT24, cc->ptr_list[1], (u_short *)&dat);
    if (i >= cc->ptr_list[2])
	return CC_STA_LIST_TIMEOUT;
    else
        return 0;
}


cc_qs16(cc)
struct cc_device *cc;
{
    register u_short mode, naf;
    register int status, len;
    int retlen;

    mode = CC_BLOCK | CC_BIT16 | CC_QSTOP;
    naf = cc->ptr_list[1];
    len = cc->ptr_list[2];
    flush_buf(cc);
    if (status = check_buf(cc, len))
        return status;
    return camac_b(getminor(cc->dev), mode, naf, len, &retlen);
}


cc_qs24(cc)
struct cc_device *cc;
{
    register u_short mode, naf;
    register int status, len;
    int retlen;

    mode = CC_BLOCK | CC_BIT24 | CC_QSTOP;
    naf = cc->ptr_list[1];
    len = cc->ptr_list[2];
    flush_buf(cc);
    if (status = check_buf(cc, len * 2))
        return status;
    return camac_b(getminor(cc->dev), mode, naf, len, &retlen);
}


cc_qs16_a(cc)
struct cc_device *cc;
{
    register u_short mode, naf;
    register int status, len;
    int retlen;

    mode = CC_BLOCK | CC_BIT16 | CC_QSTOP;
    naf = cc->ptr_list[1];
    len = cc->a_reg;
    flush_buf(cc);
    if (status = check_buf(cc, len))
        return status;
    return camac_b(getminor(cc->dev), mode, naf, len, &retlen);
}


cc_qs24_a(cc)
struct cc_device *cc;
{
    register u_short mode, naf;
    register int status, len;
    int retlen;

    mode = CC_BLOCK | CC_BIT24 | CC_QSTOP;
    naf = cc->ptr_list[1];
    len = cc->a_reg;
    flush_buf(cc);
    if (status = check_buf(cc, len * 2))
        return status;
    return camac_b(getminor(cc->dev), mode, naf, len, &retlen);
}


cc_as16(cc)
struct cc_device *cc;
{
    register u_short mode, naf;
    register int status, len;
    int retlen;

    mode = CC_BLOCK | CC_BIT16 | CC_QSCAN;
    naf = cc->ptr_list[1];
    len = cc->ptr_list[2];
    flush_buf(cc);
    if (status = check_buf(cc, len))
        return status;
    return camac_b(getminor(cc->dev), mode, naf, len, &retlen);
}


cc_as24(cc)
struct cc_device *cc;
{
    register u_short mode, naf;
    register int status, len;
    int retlen;

    mode = CC_BLOCK | CC_BIT24 | CC_QSCAN;
    naf = cc->ptr_list[1];
    len = cc->ptr_list[2];
    flush_buf(cc);
    if (status = check_buf(cc, len * 2))
        return status;
    return camac_b(getminor(cc->dev), mode, naf, len, &retlen);
}


cc_as16_a(cc)
struct cc_device *cc;
{
    register u_short mode, naf;
    register int status, len;
    int retlen;

    mode = CC_BLOCK | CC_BIT16 | CC_QSCAN;
    naf = cc->ptr_list[1];
    len = cc->a_reg;
    flush_buf(cc);
    if (status = check_buf(cc, len))
        return status;
    return camac_b(getminor(cc->dev), mode, naf, len, &retlen);
}


cc_as24_a(cc)
struct cc_device *cc;
{
    register u_short mode, naf;
    register int status, len;
    int retlen;

    mode = CC_BLOCK | CC_BIT24 | CC_QSCAN;
    naf = cc->ptr_list[1];
    len = cc->a_reg;
    flush_buf(cc);
    if (status = check_buf(cc, len * 2))
        return status;
    return camac_b(getminor(cc->dev), mode, naf, len, &retlen);
}


cc_igq16(cc)
struct cc_device *cc;
{
    register u_short mode, naf;
    register int status, len;
    int retlen;

    mode = CC_BLOCK | CC_BIT16 | CC_QIGNO;
    naf = cc->ptr_list[1];
    len = cc->ptr_list[2];
    flush_buf(cc);
    if (status = check_buf(cc, len))
        return status;
    return camac_b(getminor(cc->dev), mode, naf, len, &retlen);
}


cc_igq24(cc)
struct cc_device *cc;
{
    register u_short mode, naf;
    register int status, len;
    int retlen;

    mode = CC_BLOCK | CC_BIT24 | CC_QIGNO;
    naf = cc->ptr_list[1];
    len = cc->ptr_list[2];
    flush_buf(cc);
    if (status = check_buf(cc, len * 2))
        return status;
    return camac_b(getminor(cc->dev), mode, naf, len, &retlen);
}


cc_igq16_a(cc)
struct cc_device *cc;
{
    register u_short mode, naf;
    register int status, len;
    int retlen;

    mode = CC_BLOCK | CC_BIT16 | CC_QIGNO;
    naf = cc->ptr_list[1];
    len = cc->a_reg;
    flush_buf(cc);
    if (status = check_buf(cc, len))
        return status;
    return camac_b(getminor(cc->dev), mode, naf, len, &retlen);
}


cc_igq24_a(cc)
struct cc_device *cc;
{
    register u_short mode, naf;
    register int status, len;
    int retlen;

    mode = CC_BLOCK | CC_BIT24 | CC_QIGNO;
    naf = cc->ptr_list[1];
    len = cc->a_reg;
    flush_buf(cc);
    if (status = check_buf(cc, len * 2))
        return status;
    return camac_b(getminor(cc->dev), mode, naf, len, &retlen);
}


int cc_move_as(cc)
struct cc_device *cc;
{
    cc->s_reg = cc->a_reg;
    return 0;
}

int cc_move_at(cc)
struct cc_device *cc;
{
    cc->t_reg = cc->a_reg;
    return 0;
}

int cc_move_sa(cc)
struct cc_device *cc;
{
    cc->a_reg = cc->s_reg;
    return 0;
}

int cc_move_st(cc)
struct cc_device *cc;
{
    cc->t_reg = cc->s_reg;
    return 0;
}

int cc_move_ta(cc)
struct cc_device *cc;
{
    cc->a_reg = cc->t_reg;
    return 0;
}

int cc_move_ts(cc)
struct cc_device *cc;
{
    cc->s_reg = cc->t_reg;
    return 0;
}

int cc_move32_a(cc)
struct cc_device *cc;
{
    cc->a_reg = Tolong(1);
    return 0;
}

int cc_move32_s(cc)
struct cc_device *cc;
{
    cc->s_reg = Tolong(1);
    return 0;
}

int cc_move32_t(cc)
struct cc_device *cc;
{
    cc->t_reg = Tolong(1);
    return 0;
}

int cc_move16_a_ptr(cc)
struct cc_device *cc;
{
    register int status;

    if (status = check_buf(cc, 1))
	return status;
    *cc->ptr_kdata++ = (u_short)cc->a_reg;
    cc->len_kdata++;
    return 0;
}

int cc_move32_a_ptr(cc)
struct cc_device *cc;
{
    register int status;

    if (status = check_buf(cc, 2))
	return status;
    *cc->ptr_kdata++ = *((u_short *)&cc->a_reg);
    *cc->ptr_kdata++ = *((u_short *)&cc->a_reg + 1);
    cc->len_kdata += 2;
    return 0;
}

int cc_move16_ptr(cc)
struct cc_device *cc;
{
    register int status;

    if (status = check_buf(cc, 1))
        return status;
    *cc->ptr_kdata++ = cc->ptr_list[1];
    cc->len_kdata++;
    return 0;
}

int cc_move32_ptr(cc)
struct cc_device *cc;
{
    register int status;

    if (status = check_buf(cc, 2))
	return status;
    *cc->ptr_kdata++ = cc->ptr_list[2];
    *cc->ptr_kdata++ = cc->ptr_list[1];
    cc->len_kdata += 2;
    return 0;
}

int cc_move16_a_head(cc)
struct cc_device *cc;
{
    cc->cc_sys_status = copyout((caddr_t)((u_short *)&cc->a_reg + 1),
			    (caddr_t)&ptr_prev_head[cc->ptr_list[1]-1], 2);
    return 0;
}

int cc_move16_head_a(cc)
struct cc_device *cc;
{
    cc->a_reg = 0;
    cc->cc_sys_status = copyin((caddr_t)&ptr_prev_head[cc->ptr_list[1]-1],
			   (caddr_t)((u_short *)&cc->a_reg + 1), 2);
    return 0;
}

int cc_move16_head(cc)
struct cc_device *cc;
{
    cc->cc_sys_status = copyout((caddr_t)&cc->ptr_list[2],
			    (caddr_t)&ptr_prev_head[cc->ptr_list[1]-1], 2);
    return 0;
}

int cc_load_ptr(cc)
struct cc_device *cc;
{
    cc->a_reg = cc->len_udata + cc->len_kdata;
    return 0;
}

int cc_load_remain(cc)
struct cc_device *cc;
{
    cc->a_reg = cc->len_udata_t - (cc->len_udata + cc->len_kdata);
    return 0;
}

int cc_load_timer(cc)
struct cc_device *cc;
{
/*
    cc->a_reg = 0;
    return 0;
*/

/*** temporary ***/
    return klexe(cc, cc->klist1[0].name);
}

int cc_load_count(cc)
struct cc_device *cc;
{
    cc->a_reg = cc->event_count;
    return 0;
}

int cc_load_qx(cc)
struct cc_device *cc;
{
    cc->a_reg = cc->camac_qx;
    return 0;
}

int cc_head(cc)
struct cc_device *cc;
{
    ptr_prev_head = cc->ptr_udata;
    cc->ptr_udata += cc->ptr_list[1];
    cc->len_udata += cc->ptr_list[1];
    return 0;
}

int cc_exchg_at(cc)
struct cc_device *cc;
{
    register u_int i;

    i = cc->a_reg;
    cc->a_reg = cc->t_reg;
    cc->t_reg = i;
    return 0;
}

int cc_swap_a(cc)
struct cc_device *cc;
{
    register u_short temp;

    swap.longs = cc->a_reg;
    temp = swap.shorts[0];
    swap.shorts[0] = swap.shorts[1];
    swap.shorts[1] = temp;
    cc->a_reg = swap.longs;
    return 0;
}


int cc_bis32_at(cc)
struct cc_device *cc;
{
    cc->t_reg |= cc->a_reg;
    return 0;
}

int cc_bis32_ta(cc)
struct cc_device *cc;
{
    cc->a_reg |= cc->t_reg;
    return 0;
}

int cc_bis32_a(cc)
struct cc_device *cc;
{
    cc->a_reg |= Tolong(1);
    return 0;
}

int cc_bis32_t(cc)
struct cc_device *cc;
{
    cc->t_reg |= Tolong(1);
    return 0;
}

int cc_bic32_at(cc)
struct cc_device *cc;
{
    cc->t_reg &= ~cc->a_reg;
    return 0;
}

int cc_bic32_ta(cc)
struct cc_device *cc;
{
    cc->a_reg &= ~cc->t_reg;
    return 0;
}

int cc_bic32_a(cc)
struct cc_device *cc;
{
    cc->a_reg &= ~Tolong(1);
    return 0;
}

int cc_bic32_t(cc)
struct cc_device *cc;
{
    cc->t_reg &= ~Tolong(1);
    return 0;
}

int cc_add32_at(cc)
struct cc_device *cc;
{
    cc->t_reg += cc->a_reg;
    return 0;
}

int cc_add32_ta(cc)
struct cc_device *cc;
{
    cc->a_reg += cc->t_reg;
    return 0;
}

int cc_add32_a(cc)
struct cc_device *cc;
{
    cc->a_reg += Tolong(1);
    return 0;
}

int cc_add32_t(cc)
struct cc_device *cc;
{
    cc->t_reg += Tolong(1);
    return 0;
}

int cc_sub32_at(cc)
struct cc_device *cc;
{
    cc->t_reg -= cc->a_reg;
    return 0;
}

int cc_sub32_ta(cc)
struct cc_device *cc;
{
    cc->a_reg -= cc->t_reg;
    return 0;
}

int cc_sub32_a(cc)
struct cc_device *cc;
{
    cc->a_reg -= Tolong(1);
    return 0;
}

int cc_sub32_t(cc)
struct cc_device *cc;
{
    cc->t_reg -= Tolong(1);
    return 0;
}

int cc_mul32_at(cc)
struct cc_device *cc;
{
    cc->t_reg *= cc->a_reg;
    return 0;
}

int cc_mul32_ta(cc)
struct cc_device *cc;
{
    cc->a_reg *= cc->t_reg;
    return 0;
}

int cc_mul32_a(cc)
struct cc_device *cc;
{
    cc->a_reg *= Tolong(1);
    return 0;
}

int cc_mul32_t(cc)
struct cc_device *cc;
{
    cc->t_reg *= Tolong(1);
    return 0;
}

int cc_div32_at(cc)
struct cc_device *cc;
{
    if (cc->a_reg == 0)
        return CC_STA_LIST_ZERODIV;
    cc->t_reg /= cc->a_reg;
    return 0;
}

int cc_div32_ta(cc)
struct cc_device *cc;
{
    if (cc->t_reg == 0)
        return CC_STA_LIST_ZERODIV;
    cc->a_reg /= cc->t_reg;
    return 0;
}

int cc_div32_a(cc)
struct cc_device *cc;
{
    register int i;

    if ((i = Tolong(1)) == 0)
        return CC_STA_LIST_ZERODIV;
    cc->a_reg /= i;
    return 0;
}


int cc_div32_t(cc)
struct cc_device *cc;
{
    register int i;

    if ((i = Tolong(1)) == 0)
        return CC_STA_LIST_ZERODIV;
    cc->t_reg /= i;
    return 0;
}

int cc_waitint(cc)
struct cc_device *cc;
{
    int s;
    u_long tick, timeout;

    cc->interrupt = 0;
    cc->k->lamc = CC_INT_AUTO_CLEAR | CC_INT_ENABLE | intrpri;

    /* lock out clock */
    s = spl5();
    if (cc->interrupt & CC_INT_LAM) {
	cc->interrupt &= ~CC_INT_LAM;
 	return 0;
    }

    mutex_enter(&cc->mutex);                        /* start MUTEX */
    drv_getparm(LBOLT, &tick);
#ifdef CC_SPARC_LIST_EXTENSION
    timeout = tick + cc->a_reg * hz;
#else
    timeout = tick + CC_TIMEOUT_LAM * hz;
#endif
    /* wait interrupt */
    if (cv_timedwait(&cc->cv, &cc->mutex, timeout) == 0)
        cc->interrupt |= CC_INT_TIMEOUT;
    mutex_exit(&cc->mutex);                         /* end MUTEX */
    
    splx(s);

    if (cc->interrupt & CC_INT_LAM) {
	cc->interrupt &= ~CC_INT_LAM;
	return 0;
    }
    else {
        return CC_STA_LIST_TIMEOUT;
    }
}


int cc_waitevent(cc)
struct cc_device *cc;
{
    int s;
    u_long tick, timeout;

    cc->interrupt = 0;
    cc->k->lamc = CC_INT_AUTO_CLEAR | CC_INT_ENABLE | intrpri;

    /* lock out clock */
    s = spl5();
    if (cc->interrupt & CC_INT_LAM) {
	cc->interrupt &= ~CC_INT_LAM;
	cc->event_count++;
 	return 0;
    }

    mutex_enter(&cc->mutex);                        /* start MUTEX */
    drv_getparm(LBOLT, &tick);
#ifdef CC_SPARC_LIST_EXTENSION
    timeout = tick + cc->a_reg * hz;
#else
    timeout = tick + CC_TIMEOUT_LAM * hz;
#endif
    /* wait interrupt */
    if (cv_timedwait(&cc->cv, &cc->mutex, timeout) == 0)
        cc->interrupt |= CC_INT_TIMEOUT;
    mutex_exit(&cc->mutex);                         /* end MUTEX */
    
    splx(s);

    if (cc->interrupt & CC_INT_LAM) {
	cc->interrupt &= ~CC_INT_LAM;
	cc->event_count++;
	return 0;
    }
    else {
        return CC_STA_LIST_TIMEOUT;
    }
}


int cc_br(cc)
struct cc_device *cc;
{
    cc->ptr_list = cc->list_area + cc->ptr_list[1] - 1;
    cc->branch_flag = 1;
    return 0;
}

int cc_beq(cc)
struct cc_device *cc;
{
    if (cc->t_reg == Tolong(2)) {
	cc->ptr_list = cc->list_area + cc->ptr_list[1] - 1;
	cc->branch_flag = 1;
    }
    return 0;
}

int cc_bne(cc)
struct cc_device *cc;
{
    if (cc->t_reg != Tolong(2)) {
	cc->ptr_list = cc->list_area + cc->ptr_list[1] - 1;
	cc->branch_flag = 1;
    }
    return 0;
}

int cc_bgt(cc)
struct cc_device *cc;
{
    if (cc->t_reg > Tolong(2)) {
	cc->ptr_list = cc->list_area + cc->ptr_list[1] - 1;
	cc->branch_flag = 1;
    }
    return 0;
}

int cc_bge(cc)
struct cc_device *cc;
{
    if (cc->t_reg >= Tolong(2)) {
	cc->ptr_list = cc->list_area + cc->ptr_list[1] - 1;
	cc->branch_flag = 1;
    }
    return 0;
}

int cc_blt(cc)
struct cc_device *cc;
{
    if (cc->t_reg < Tolong(2)) {
	cc->ptr_list = cc->list_area + cc->ptr_list[1] - 1;
	cc->branch_flag = 1;
    }
    return 0;
}

int cc_ble(cc)
struct cc_device *cc;
{
    if (cc->t_reg <= Tolong(2)) {
	cc->ptr_list = cc->list_area + cc->ptr_list[1] - 1;
	cc->branch_flag = 1;
    }
    return 0;
}

int cc_bbc(cc)
struct cc_device *cc;
{
    if (!(cc->t_reg & (1 << Tolong(2)))) {
	cc->ptr_list = cc->list_area + cc->ptr_list[1] - 1;
	cc->branch_flag = 1;
    }
    return 0;
}

int cc_bbs(cc)
struct cc_device *cc;
{
    if (cc->t_reg & (1 <<  Tolong(2))) {
	cc->ptr_list = cc->list_area + cc->ptr_list[1] - 1;
	cc->branch_flag = 1;
    }
    return 0;
}


/********************************************************************
 * klist.c         20-FEB-1992            Y.Takeuchi (T.I.T.)       *
 *                                                                  *
 * K2917's list processing                                          *
 * for SPARC cc device driver  K2917  K3922                         *
 *                                                                  *
 * 26-NOV-1992 Y.Takeuchi    modify                                 *
 *  5-MAR-1993 Y.Takeuchi    modify for unified UNIX camlib         *
 ********************************************************************/

kladd(cc, name)
struct cc_device *cc;
char *name;
{
/*  register struct K_REG *k = cc->k; */
    register int i, code;

    cc->status = 0;

    /* same name check */
    for (i = 0; i < cc->ptr_dir; i++)
	if (strcmp(name, cc->klist1[i].name) == 0)
	    return CC_STA_KLIST_SAMENAME;

    if (cc->ptr_dir >= CC_KLIST_MAX_DIR)
        return CC_STA_KLIST_MEMORYFULL;

    /* add */
    cc->k->cma = cc->ptr_cma;
    cc->klist1[cc->ptr_dir].cma = cc->ptr_cma;
    cc->klist1[cc->ptr_dir].wc = 0;

    cc->ptr_list = cc->list_area;
    while (cc->status == 0) {
	if (cc->ptr_cma >= 0xFFFF - 4)
	    return CC_STA_KLIST_MEMORYFULL;
	code = 0x00FF & cc->ptr_list[0] + 0x60;
	instruction_call = instruction[code];
	cc->status = (*instruction_call)(cc);
	next_instruction(cc);
    }

    strcpy(cc->klist1[cc->ptr_dir].name, name);
    cc->ptr_dir++;

    return 0;
}


kldel(name)
char *name;
{
struct cc_device *cc;
    register int i;

    /* same name check */
    for (i = 0; i < cc->ptr_dir; i++)
	if (strcmp(name, cc->klist1[i].name) == 0) {
	    cc->klist1[i].name[0] = '\0';
	    return 0;
	}

    return CC_STA_KLIST_NONAME;
}


klexe(cc, name)
struct cc_device *cc;
char *name;
{
    register dev_t dev = cc->dev;
    register struct uio *uio = cc->uio;
    register int *retlen = (int *)cc->ptr_udata;
    register int i, status;

    cc->klist = CC_KLIST_YES;

    /* same name check */
    for (i = 0; i < cc->ptr_dir; i++)
	if (strcmp(name, cc->klist1[i].name) == 0)
	    break;
    if (i >= cc->ptr_dir)
        return CC_STA_KLIST_NONAME;

    cc->now_cma = cc->klist1[i].cma;
    cc->now_wc  = cc->klist1[i].wc;

    /* first 2words for retlen */
    uio->uio_iov->iov_base = (caddr_t)(cc->ptr_udata + 2);
    uio->uio_iov->iov_len  = cc->klist1[i].wc * 2;
    uio->uio_iovcnt = 1;
    uio->uio_offset = 0;

    /* Kinetic 2917 list is now READ ONLY. */
    status = physio(cc_strategy, (struct buf *)NULL, dev, B_READ,
		    cc_minphys, uio);
    *retlen = cc->retlen;

    return status;
}


klsave(cc, data)
struct cc_device *cc;
u_short *data;
{
/*  register struct K_REG *k = cc->k; */
    register int i, j, ptr = 1;
    register u_short save;

    for (i = 0; i < cc->ptr_dir; i++)
	if (cc->klist1[i].name[0] != '\0') {

	    /* save list name */
	    j = 0;
	    while (cc->klist1[i].name[j] != '\0')
	        data[ptr++] = cc->klist1[i].name[j++];
	    data[ptr++] = '\0';

	    /* save data word counts */
	    data[ptr++] = cc->klist1[i].wc;

	    /* save list */
	    cc->k->cma = cc->klist1[i].cma;
	    while ((save = cc->k->cmr) != CC_HALT)
	        data[ptr++] = save;
	    data[ptr++] = CC_HALT;
	}

    data[0] = ptr - 1;

    return 0;
}


klload(cc, data)
struct cc_device *cc;
u_short *data;
{
/*  register struct K_REG *k = cc->k; */
    register int j, ptr = 1;

    while (ptr <= data[0]) {
	/* load list name */
	j = 0;
	while (data[ptr] != '\0')
	    cc->klist1[cc->ptr_dir].name[j++] = data[ptr++];
	cc->klist1[cc->ptr_dir].name[j] = data[ptr++];
	
	/* load data word counts */
	  cc->klist1[cc->ptr_dir].wc = data[ptr++];
	
	/* load list */
	cc->k->cma = cc->ptr_cma; 
	cc->klist1[cc->ptr_dir].cma = cc->ptr_cma;
	while (data[ptr] != CC_HALT) {
	    cc->k->cmr = data[ptr++];
	    cc->ptr_cma++;
	}
	cc->k->cmr = data[ptr++];
	cc->ptr_cma++;

	cc->ptr_dir++;
    }

    return 0;
}




/************************************************
 * K2917 list                                   *
 ************************************************/
kl_stop(cc)
struct cc_device *cc;
{
/*  register struct K_REG *k = cc->k; */

    cc->k->cmr = CC_HALT;
    return CC_STA_LIST_STOP;
}

kl_read16(cc)
struct cc_device *cc;
{
/*  register struct K_REG *k = cc->k; */

    cc->k->cmr = CC_AD | CC_BIT16 | (cc->cur_crate << 8);
    cc->k->cmr = cc->ptr_list[1];
    cc->klist1[cc->ptr_dir].wc += 1;
    cc->ptr_cma += 2;
    return 0;
}

kl_read24(cc)
struct cc_device *cc;
{
/*  register struct K_REG *k = cc->k; */

    cc->k->cmr = CC_AD | CC_BIT24 | (cc->cur_crate << 8);
    cc->k->cmr = cc->ptr_list[1];
    cc->klist1[cc->ptr_dir].wc += 2;
    cc->ptr_cma += 2;
    return 0;
}

kl_write16(cc)
struct cc_device *cc;
{
/*  register struct K_REG *k = cc->k; */

    cc->k->cmr = CC_AD | CC_INLINE | CC_BIT16 | (cc->cur_crate << 8);
    cc->k->cmr = cc->ptr_list[1];
    cc->k->cmr = cc->ptr_list[2];
    cc->k->cmr = 0;
    cc->ptr_cma += 4;
    return 0;
}

kl_write24(cc)
struct cc_device *cc;
{
/*  register struct K_REG *k = cc->k; */

    cc->k->cmr = CC_AD | CC_INLINE | CC_BIT24 | (cc->cur_crate << 8);
    cc->k->cmr = cc->ptr_list[1];
    cc->k->cmr = cc->ptr_list[2];
    cc->k->cmr = cc->ptr_list[3];
    cc->ptr_cma += 4;
    return 0;
}

kl_ndt(cc)
struct cc_device *cc;
{
/*  register struct K_REG *k = cc->k; */

    cc->k->cmr = CC_AD | CC_BIT16 | (cc->cur_crate << 8);
    cc->k->cmr = cc->ptr_list[1];
    cc->ptr_cma += 2;
    return 0;
}

kl_qs16(cc)
struct cc_device *cc;
{
/*  register struct K_REG *k = cc->k; */

    cc->k->cmr = CC_AD | CC_BLOCK | CC_BIT16 | CC_QSTOP | (cc->cur_crate << 8);
    cc->k->cmr = cc->ptr_list[1];
    cc->k->cmr = - cc->ptr_list[2];
    cc->k->cmr = 0xFFFF;
    cc->klist1[cc->ptr_dir].wc += cc->ptr_list[2];
    cc->ptr_cma += 4;
    return 0;
}

kl_qs24(cc)
struct cc_device *cc;
{
/*  register struct K_REG *k = cc->k; */

    cc->k->cmr = CC_AD | CC_BLOCK | CC_BIT24 | CC_QSTOP | (cc->cur_crate << 8);
    cc->k->cmr = cc->ptr_list[1];
    cc->k->cmr = - cc->ptr_list[2] * 2;
    cc->k->cmr = 0xFFFF;
    cc->klist1[cc->ptr_dir].wc += cc->ptr_list[2] * 2;
    cc->ptr_cma += 4;
    return 0;
}

kl_as16(cc)
struct cc_device *cc;
{
/*  register struct K_REG *k = cc->k; */

    cc->k->cmr = CC_AD | CC_BLOCK | CC_BIT16 | CC_QSCAN | (cc->cur_crate << 8);
    cc->k->cmr = cc->ptr_list[1];
    cc->k->cmr = - cc->ptr_list[2];
    cc->k->cmr = 0xFFFF;
    cc->klist1[cc->ptr_dir].wc += cc->ptr_list[2];
    cc->ptr_cma += 4;
    return 0;
}

kl_as24(cc)
struct cc_device *cc;
{
/*  register struct K_REG *k = cc->k; */

    cc->k->cmr = CC_AD | CC_BLOCK | CC_BIT24 | CC_QSCAN | (cc->cur_crate << 8);
    cc->k->cmr = cc->ptr_list[1];
    cc->k->cmr = - cc->ptr_list[2] * 2;
    cc->k->cmr = 0xFFFF;
    cc->klist1[cc->ptr_dir].wc += cc->ptr_list[2] * 2;
    cc->ptr_cma += 4;
    return 0;
}

kl_igq16(cc)
struct cc_device *cc;
{
/*  register struct K_REG *k = cc->k; */

    cc->k->cmr = CC_AD | CC_BLOCK | CC_BIT16 | CC_QIGNO | (cc->cur_crate << 8);
    cc->k->cmr = cc->ptr_list[1];
    cc->k->cmr = - cc->ptr_list[2];
    cc->k->cmr = 0xFFFF;
    cc->klist1[cc->ptr_dir].wc += cc->ptr_list[2];
    cc->ptr_cma += 4;
    return 0;
}

kl_igq24(cc)
struct cc_device *cc;
{
/*  register struct K_REG *k = cc->k; */

    cc->k->cmr = CC_AD | CC_BLOCK | CC_BIT24 | CC_QIGNO | (cc->cur_crate << 8);
    cc->k->cmr = cc->ptr_list[1];
    cc->k->cmr = - cc->ptr_list[2] * 2;
    cc->k->cmr = 0xFFFF;
    cc->klist1[cc->ptr_dir].wc += cc->ptr_list[2] * 2;
    cc->ptr_cma += 4;
    return 0;
}

/********************************************************************
 * kcam_ss.c       22-SEP-1991            Y.Takeuchi                *
 *                                                                  *
 * Slave Single Transfer                                            *
 * for SPARC cc device driver  K2917  K3922                         *
 *                                                                  *
 * 13-JUL-1992 Y.Takeuchi    modify for new-K2917 board             *
 *  5-MAR-1993 Y.Takeuchi    modify for unified UNIX camlib         *
 ********************************************************************/
static int
camac_s(int unit, u_short mode, u_short naf, u_short *dat)
{
    register struct cc_device *cc;
/*  register struct K_REG *k = cc->k; */
    register int counter;

    cc = ddi_get_soft_state(cc_state,unit);

    cc->status = 0;

/* by E.Inoue */
/*    k->csr = CC_RST;                  */       /* K2917 Reset */

/* E.Inoue */
    cc->k->lamc = (u_short)CC_INT_AUTO_CLEAR;
    cc->k->donc = (u_short)CC_INT_AUTO_CLEAR;
    cc->k->empc = (u_short)CC_INT_AUTO_CLEAR;
    cc->k->aboc = (u_short)CC_INT_AUTO_CLEAR;
/* end */

    cc->k->cma = CC_CMA_INIT;                    /* Initialize memory pointer */
    cc->k->cmr = mode | (cc->cur_crate << 8); /* Write command list */
    cc->k->cmr = naf;
    cc->k->cmr = CC_HALT;
    cc->k->cma = CC_CMA_INIT;                    /* Reset memory pointer */

    counter = 0;
    switch (naf & 0x0018) {
      case 0x0000:                        /* CAMAC read */
/* E.Inoue: from 
    cmn_err(CE_NOTE," camac_s: debug step02. --- camac read");
 E.Inoue: end */
/* E.Inoue 
    cc->k->lamc = (u_short)CC_INT_AUTO_CLEAR;
    cc->k->donc = (u_short)CC_INT_AUTO_CLEAR;
    cc->k->empc = (u_short)CC_INT_AUTO_CLEAR;
    cc->k->aboc = (u_short)CC_INT_AUTO_CLEAR;
 end */

	cc->k->csr &= ~CC_WRITE;
	cc->k->csr |= CC_GO;                     /* Go! */
	while ((cc->k->csr & (CC_RDY|CC_ERR)) == 0 && counter < CC_TIMEOUT_SINGLE)
	    counter++;
	if ((cc->k->csr & CC_RDY) != 0) {
	    if ((mode & CC_BIT16) == 0) {
		*dat = cc->k->dhr & 0x00FF;
		*(dat + 1) = cc->k->dlr;
	    }
	    else
		*dat = cc->k->dlr;
	}
	break;

      case 0x0010:                        /* CAMAC write */

/* E.Inoue 
    cc->k->lamc = (u_short)CC_INT_AUTO_CLEAR;
    cc->k->donc = (u_short)CC_INT_AUTO_CLEAR;
    cc->k->empc = (u_short)CC_INT_AUTO_CLEAR;
    cc->k->aboc = (u_short)CC_INT_AUTO_CLEAR;
 end */

	cc->k->csr |= CC_WRITE;
	cc->k->csr |= CC_GO;                     /* Go! */
	while ((cc->k->csr & (CC_RDY|CC_ERR)) == 0 && counter < CC_TIMEOUT_SINGLE)
	    counter++;
	if ((cc->k->csr & CC_RDY) != 0) {
	    if ((mode & CC_BIT16) == 0) {
		cc->k->dhr = *dat;
		cc->k->dlr = *(dat + 1);
	    }
	    else {
		cc->k->dlr = *dat;
   		}
	}
	break;

      default:                            /* NDT */
	cc->k->csr |= CC_GO;                     /* Go! */
	break;
    }

    while ((cc->k->csr & (CC_DONE | CC_ERR)) == 0 && counter < CC_TIMEOUT_SINGLE)
        counter++;

    cc->camac_qx = cc->k->csr;

    if (counter >= CC_TIMEOUT_SINGLE) {
        cc->status = CC_STA_SINGLE_TIMEOUT;
    }

    return cc->status;
}


/********************************************************************
 * kcam_mb.c       19-FEB-1992            Y.Takeuchi (T.I.T.)       *
 *                                                                  *
 * FOR K2917 list processing                                        *
 * Master Block Transfer to SUN memory (DVMA+DMA)                   *
 * for SPARC cc device driver  K2917  K3922                         *
 *                                                                  *
 * 13-JUL-1992 Y.Takeuchi    use "done interrupt" to detect EOT     *
 * 30-NOV-1992 Y.Takeuchi                                           *
 *  5-MAR-1993 Y.Takeuchi    modify for unified UNIX camlib         *
 ********************************************************************/
static void cc_minphys(struct buf *bp)
{
    if (bp->b_bcount > CC_MINPHYS_SIZE)
        bp->b_bcount = CC_MINPHYS_SIZE;

/* E.Inoue: begin */
	minphys(bp);
/* E.Inoue: end */

    return;
}

/* change logichouse */
/* setup DMA limit */
/*
static ddi_dma_lim_t dma_lim = {
    0,
    0xffffffff,
    0xffff,
    0x2,
    0x2,
    1024
};
*/
/* end */

static int cc_strategy(struct buf *bp)
{
/*  register struct cc_device *cc = &ccdevice[0]; */
    register struct cc_device *cc;
/*  register struct K_REG *k = cc->k; */
    register u_long dma_addr;
/*
    register mode = cc->mode;
    register naf  = cc->naf;
    register int wc = (mode & CC_BIT16) ? cc->len : cc->len * 2;
*/
    register mode;
    register naf;
    register int wc;
    register u_int flags;
    minor_t unit;

    cc = ddi_get_soft_state(cc_state, getminor(bp->b_edev));

    mode = cc->mode;
    naf  = cc->naf;
    wc = (mode & CC_BIT16) ? cc->len : cc->len * 2;

    /* check DMA mode and set flags */
    switch (cc->naf & 0x0018) {
        case 0x0000:                        /* CAMAC read */
            flags = DDI_DMA_READ;
            break;
        case 0x0010:                        /* CAMAC write */
            flags = DDI_DMA_WRITE;
            break;
        default:
            cc->status = CC_STA_BLOCK_INVFUNC;
        return 0;
    }

    mutex_enter(&cc->mutex);

    while(cc->executing_dma_flag){
        cv_wait(&cc->cv,&cc->mutex);
    }

    mutex_exit(&cc->mutex);


/* change logichouse */
    /* setup DMA bufer */

    if((ddi_dma_buf_bind_handle(cc->handle, bp ,
        flags | DDI_DMA_CONSISTENT, DDI_DMA_SLEEP, (caddr_t)0,
        &cc->dma_cookie, &cc->ccount) != DDI_DMA_MAPPED)){
        cmn_err(CE_WARN,"cc_strategy: ddi_dma_buf_bind_handle failed");
        bp->b_flags |= B_ERROR;
        bp->b_error = EIO;
        return bp->b_flags;
    }

/*
    if (ddi_dma_buf_setup(cc->dip, bp, flags, DDI_DMA_SLEEP, NULL,
     &dma_lim, &cc->handle) != DDI_DMA_MAPPED) {
         bp->b_error |= EIO;
         bp->b_flags |= B_ERROR;
         return bp->b_flags;
    }
*/
    /* get a virtual address to hand to our device */
/*
    if (ddi_dma_htoc(cc->handle, NULL, &cc->dma_cookie) != DDI_SUCCESS) {
        bp->b_error |= EIO;
        bp->b_flags |= B_ERROR;
        return bp->b_flags;
    }
*/
/* end */

    /* set up variables */
    cc->bp = bp;

    dma_addr = cc->dma_cookie.dmac_address;
    /* count = cc->dma_cookie.dmac_size; */

    cc->retlen = 0;
/*    k->csr = CC_RST;  */                        /* K2917 Reset */

    if (cc->klist == CC_KLIST_NO) {
        cc->k->cma = CC_CMA_INIT;               /* Initialize memory pointer */
        cc->k->cmr = mode | (cc->cur_crate << 8);
        cc->k->cmr = naf;
        cc->k->cmr = -(cc->len & 0xFFFF);         /* Max len = 1MWord */
        cc->k->cmr = 0xFFFF;
        cc->k->cmr = CC_HALT;
        cc->k->cma = CC_CMA_INIT;                 /* Reset memory pointer */
        cc->k->maclo = dma_addr & 0xFFFF;         /* Set DMA base address */
        cc->k->machi = dma_addr >> 16;
        cc->k->amr = CC_AMR_INIT;                 /* Set VME AM code */
        cc->k->mtc = wc;
        cc->k->cser = CC_DMA_RESET;               /* DMA reset */

 switch (cc->naf & 0x0018) {
     case 0x0000:                        /* CAMAC read */
         cc->k->docr = CC_DOCR_INIT | CC_DMA_READ;
         cc->k->sccr = CC_DMA_START;
         cc->k->csr |= CC_DMA;                 /* DMA mode */
         cc->k->csr &= ~CC_WRITE;
         break;

     case 0x0010:                        /* CAMAC write */
         cc->k->docr = CC_DOCR_INIT | CC_DMA_WRITE;
         cc->k->sccr = CC_DMA_START;
         cc->k->csr |= CC_DMA;                 /* DMA mode */
         cc->k->csr |= CC_WRITE;
         break;

     default:                            /* ERROR */
         cc->status = CC_STA_BLOCK_INVFUNC;
         return 0;
 }
    }
    /* Kinetic list-processing */
    else {
         cc->k->cma = cc->now_cma;                    /* Initialize memory pointer
*/
         cc->k->maclo = dma_addr & 0xFFFF;        /* Set DMA base address */
         cc->k->machi = dma_addr >> 16;
         cc->k->amr = CC_AMR_INIT;                /* Set VME AM code */
         cc->k->mtc = cc->now_wc;
         cc->k->cser = CC_DMA_RESET;              /* DMA reset */

         cc->k->docr = CC_DOCR_INIT | CC_DMA_READ;
         cc->k->sccr = CC_DMA_START;
         cc->k->csr |= CC_DMA;                    /* DMA mode */
         cc->k->csr &= ~CC_WRITE;
    }

    /* start DMA */
    cc->executing_dma_flag = 1;
    cc->timeout_id = timeout(cc_timeout, (caddr_t)&unit, CC_TIMEOUT_DMA * hz);
    cc->k->csr |= CC_GO;                         /* Go! */
    cc->k->donc = CC_INT_AUTO_CLEAR | CC_INT_ENABLE | intrpri;

    /* wait at physio() */

    return 0;
}


static int
camac_b(int unit, u_short mode, u_short naf, int len, int *retlen)
{
    register struct cc_device *cc;
    register dev_t dev;
/*  register struct K_REG *k = cc->k; */
/*    register struct uio *uio = cc->uio; */
    register struct uio *uio;

    cc = ddi_get_soft_state(cc_state,unit);

    dev = cc->dev;
    uio = cc->uio;

    cc->klist = CC_KLIST_NO;

    cc->mode = mode;
    cc->naf = naf;
    cc->len = len;

    /* set uio for DMA */
    uio->uio_iov->iov_base = (caddr_t)cc->ptr_udata;
    uio->uio_iov->iov_len  = (mode & CC_BIT16) ? len * 2: len * 4;
    uio->uio_iovcnt = 1;
    uio->uio_segflg = UIO_USERSPACE;
    uio->uio_offset = 0;

/* */
/*
    uprintf(" in: 0x%x %d %d %d %d %d\n",
	    uio->uio_iov->iov_base,
	    uio->uio_iov->iov_len,
	    uio->uio_iovcnt,
	    uio->uio_resid,
	    uio->uio_segflg,
	    uio->uio_offset);
*/
/*
    cmn_err(CE_NOTE," in: 0x%x %d %d %d %d %d",
            uio->uio_iov->iov_base,
            uio->uio_iov->iov_len,
            uio->uio_iovcnt,
            uio->uio_resid,
            uio->uio_segflg,
            uio->uio_offset);
*/
/* */

    /* execute DMA transfer */ 
    switch (cc->naf & 0x0018) {
      case 0x0000:                        /* CAMAC read */
        physio(cc_strategy, NULL, dev, B_READ, cc_minphys, uio);
	break;
      case 0x0010:                        /* CAMAC write */
        physio(cc_strategy, NULL, dev, B_WRITE, cc_minphys, uio);
	break;
      default:
	cc->status = CC_STA_BLOCK_INVFUNC;
	return 0;
    }

    cc->k->donc = CC_INT_AUTO_CLEAR | intrpri;
    cc->camac_qx = cc->k->csr;

/*  cc->cc_sys_status = cc->bp->b_error; */

    /* check timeout */
    if (cc->interrupt & CC_INT_TIMEOUT ) {
	cc->interrupt &= ~CC_INT_TIMEOUT;
	cc->status = CC_STA_BLOCK_TIMEOUT;
	*retlen = 0;
	return cc->status;
    }
/* */
/*
    uprintf("out: %d %d %d %d %d\n",
	    uio->uio_iov->iov_len,
	    uio->uio_iovcnt,
	    uio->uio_resid,
	    uio->uio_segflg,
	    uio->uio_offset);
*/
/*
    cmn_err(CE_NOTE,"out: %d %d %d %d %d",
            uio->uio_iov->iov_len,
            uio->uio_iovcnt,
            uio->uio_resid,
            uio->uio_segflg,
            uio->uio_offset);
*/
/* */
    /* Q-STOP mode or no error */ 
    if ((mode & CC_QSCAN) == 0 || (cc->k->csr & CC_ERR) == 0) {
	if ((mode & CC_BIT16) == 0) {
	    cc->retlen = cc->len - cc->k->mtc / 2; /* retlen=number of transfer */
	    if ((mode & CC_QSCAN) == 0 && (cc->k->csr & CC_ERR) != 0) /* Q-STOP */
	      cc->retlen = cc->retlen - 2;     /* delete the data of Q=0 */
	    cc->ptr_udata += cc->retlen * 2;   /* ptr(2bytes),retlen(4bytes) */
	    cc->len_udata += cc->retlen * 2;
	}
	else {
	    cc->retlen = cc->len - cc->k->mtc;
	    if ((mode & CC_QSCAN) == 0 && (cc->k->csr & CC_ERR) != 0) /* Q-STOP */
	      cc->retlen--;                    /* delete the data of Q=0 */
	    cc->ptr_udata += cc->retlen;       /* ptr(2bytes),retlen(2bytes) */
	    cc->len_udata += cc->retlen;
	}
    }

    /* return */
    *retlen = cc->retlen;
    return cc->status;
}
